# This file contains all implementation of the Nitrogen Estimation Methods that were used in literature

from turtle import shape
import numpy as np

def M1(img):
    """
        R: average red component.
    """
    return np.nanmean(img[:,:,0])


def M2(img):
    """
        G: average green component.
    """

    return np.nanmean(img[:,:,1]) 


def M3(img):
    """
        B: average blue component.
    """

    return np.nanmean(img[:,:,2])


def M4(img):
    """
        R / (R + G + B)
    """

    return np.nanmean(img[:,:,0] / (img[:,:,0].astype(int) + img[:,:,1] + img[:,:,2]))

def M5(img):
    """
        G / (R + G + B)
    """

    return np.nanmean(img[:,:,1] / (img[:,:,0].astype(int) + img[:,:,1] + img[:,:,2]))


def M6(img):
    """
        B / (R + G + B)
    """

    return np.nanmean(img[:,:,2] / (img[:,:,0].astype(int) + img[:,:,1] + img[:,:,2]))


def M7(img):
    """
        R - G
    """
    
    return np.nanmean(abs(img[:,:,0].astype(int) - img[:,:,1]))


def M8(img):
    """
        R - B
    """

    return np.nanmean(abs(img[:,:,0].astype(int) - img[:,:,2]))


def M9(img):
    """
        G - B
    """

    return np.nanmean(abs(img[:,:,1].astype(int) - img[:,:,2]))


def M10(img):
    """
        (R - G) / (R + G)
    """

    return np.nanmean(abs(img[:,:,0].astype(int) - img[:,:,1]) / (img[:,:,0].astype(int) + img[:,:,1]))


def M11(img):
    """
        (R - B) / (R + B)
    """

    return np.nanmean(abs(img[:,:,0].astype(int) - img[:,:,2]) / (img[:,:,0].astype(int) + img[:,:,2]))

def M12(img):
    """
        (G - B) / (G + B)
    """

    return np.nanmean(abs(img[:,:,1].astype(int) - img[:,:,2]) / (img[:,:,1].astype(int) + img[:,:,2]))


def M13(img):
    """
        (R - G) /(R + G + B)
    """

    return np.nanmean(abs(img[:,:,0].astype(int) - img[:,:,1]) / (img[:,:,0].astype(int) + img[:,:,1] + img[:,:,2]))


def M14(img):
    """
        (R - B) /(R + G + B)
    """

    return np.nanmean(abs(img[:,:,0].astype(int) - img[:,:,2]) / (img[:,:,0].astype(int) + img[:,:,1] + img[:,:,2]))


def M15(img):
    """
        (G - B) /(R + G + B)
    """

    return np.nanmean(abs(img[:,:,1].astype(int) - img[:,:,2]) / (img[:,:,0].astype(int) + img[:,:,1] + img[:,:,2]))


def M16(img):
    """
        (2 * R * (G - B)) / (G + B)
    """

    return np.nanmean((2*img[:,:,0]*abs(img[:,:,1].astype(int) - img[:,:,2])) / (img[:,:,1].astype(int) + img[:,:,2]))


def M17(img):
    """
        (2 * G * (R - B)) / (R + B)
    """

    return np.nanmean((2*img[:,:,1]*abs(img[:,:,0].astype(int) - img[:,:,2])) / (img[:,:,0].astype(int) + img[:,:,2]))




# This module contains different sorting algorithms code

def bubbleSort(arr):
    n = len(arr)

    # Traverse through all array elements
    for i in range(n):

        # Last i elements are already in place
        for j in range(0, n-i-1):

            # traverse the array from 0 to n-i-1
            # Swap if the element found is greater
            # than the next element
            if arr[j] > arr[j+1]:
                arr[j], arr[j+1] = arr[j+1], arr[j]



# Function to do insertion sort
def insertionSort(arr):

    # Traverse through 1 to len(arr)
    for i in range(1, len(arr)):

        key = arr[i]

        # Move elements of arr[0..i-1], that are
        # greater than key, to one position ahead
        # of their current position
        j = i-1
        while j >= 0 and key < arr[j] :
                arr[j + 1] = arr[j]
                j -= 1
        arr[j + 1] = key

def selectionSort(arr):
    # Traverse through all array elements
    for i in range(len(arr)):

        # Find the minimum element in remaining
        # unsorted array
        min_idx = i
        for j in range(i+1, len(arr)):
            if arr[min_idx] > arr[j]:
                min_idx = j

        # Swap the found minimum element with
        # the first element
        arr[i], arr[min_idx] = arr[min_idx], arr[i]

def mergeSort(arr):
    if len(arr) > 1:

         # Finding the mid of the array
        mid = len(arr)//2

        # Dividing the array elements
        L = arr[:mid]

        # into 2 halves
        R = arr[mid:]

        # Sorting the first half
        mergeSort(L)

        # Sorting the second half
        mergeSort(R)

        i = j = k = 0

        # Copy data to temp arrays L[] and R[]
        while i < len(L) and j < len(R):
            if L[i] <= R[j]:
                arr[k] = L[i]
                i += 1
            else:
                arr[k] = R[j]
                j += 1
            k += 1

        # Checking if any element was left
        while i < len(L):
            arr[k] = L[i]
            i += 1
            k += 1

        while j < len(R):
            arr[k] = R[j]
            j += 1
            k += 1

# To heapify subtree rooted at index i.
# n is size of heap


def heapify(arr, N, i):
    largest = i  # Initialize largest as root
    l = 2 * i + 1     # left = 2*i + 1
    r = 2 * i + 2     # right = 2*i + 2

    # See if left child of root exists and is
    # greater than root
    if l < N and arr[largest] < arr[l]:
        largest = l

    # See if right child of root exists and is
    # greater than root
    if r < N and arr[largest] < arr[r]:
        largest = r

    # Change root, if needed
    if largest != i:
        arr[i], arr[largest] = arr[largest], arr[i]  # swap

        # Heapify the root.
        heapify(arr, N, largest)

# The main function to sort an array of given size


def heapSort(arr):
    N = len(arr)

    # Build a maxheap.
    for i in range(N//2 - 1, -1, -1):
        heapify(arr, N, i)

    # One by one extract elements
    for i in range(N-1, 0, -1):
        arr[i], arr[0] = arr[0], arr[i]  # swap
        heapify(arr, i, 0)

# Function to find the partition position
def partition(array, low, high):

    # Choose the rightmost element as pivot
    pivot = array[high]

    # Pointer for greater element
    i = low - 1

    # Traverse through all elements
    # compare each element with pivot
    for j in range(low, high):
        if array[j] <= pivot:
            # If element smaller than pivot is found
            # swap it with the greater element pointed by i
            i = i + 1

            # Swapping element at i with element at j
            (array[i], array[j]) = (array[j], array[i])

    # Swap the pivot element with
    # e greater element specified by i
    (array[i + 1], array[high]) = (array[high], array[i + 1])

    # Return the position from where partition is done
    return i + 1

# Function to perform quicksort


def quickSort(array, low, high):
    if low < high:

        # Find pivot element such that
        # element smaller than pivot are on the left
        # element greater than pivot are on the right
        pi = partition(array, low, high)

        # Recursive call on the left of pivot
        quickSort(array, low, pi - 1)

        # Recursive call on the right of pivot
        quickSort(array, pi + 1, high)


def countSort(arr):

    # The output character array that will have sorted arr
    output = [0 for i in range(len(arr))]

    # Create a count array to store count of individual
    # characters and initialize count array as 0
    count = [0 for i in range(256)]

    # For storing the resulting answer since the
    # string is immutable
    ans = ["" for _ in arr]

    # Store count of each character
    for i in arr:
        count[i] += 1

    # Change count[i] so that count[i] now contains actual
    # position of this character in output array
    for i in range(256):
        count[i] += count[i-1]

    # Build the output character array
    for i in range(len(arr)):
        output[count[arr[i]]-1] = arr[i]
        count[arr[i]] -= 1

    # Copy the output array to arr, so that arr now
    # contains sorted characters
    for i in range(len(arr)):
        arr[i] = output[i]
    return ans



# This package encrypt and decrypt text files

def encryption(file):
    """ This function encrypts the text file
    """
    content = ""
    # Do text encryption
    try:
        with open(file, "rt") as fd:
            content = fd.read()
    except Exception as e:
        print(e)

    enc_content = ""

    for ch in content:
        if ch != '\n':
            enc_content += (chr(ord(ch) + 3))
        else:
            enc_content += ch

    try:
        with open(file, "wt") as fd:
            fd.write(enc_content)
            print(file, "encrypted successfully")
    except Exception as e:
        print(e)


def decryption(file):
    """ This function decrypts the text file
    """
    content = ""
    # Do text encryption
    try:
        with open(file, "rt") as fd:
            content = fd.read()
    except Exception as e:
        print(e)

    dec_content = ""

    for ch in content:
        if ch != "\n":
            dec_content += (chr(ord(ch) - 3))
        else:
            dec_content += ch

    try:
        with open(file, "wt") as fd:
            fd.write(dec_content)
            print(file, "decrypted successfully")
    except Exception as e:
        print(e)
